#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.

# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import json
import mock
import sys
import unittest

from testfixtures import LogCapture

from ci_cli.tests import (
    capture_stdout,
    FakeResponse,
    MainScriptTestCase,
)
from ci_cli.ticket import (
    new_ticket,
    update_ticket,
    MismatchSuitesError,
    QueueingTicketError,
    SourceAlreadyUploadedError,
    SourcePackageUpload,
    SourcePackageUploadError,
    SubTicket,
    Ticket,
)
from ci_cli import utils
from ci_utils.testing import (
    TestCaseWithGnupg,
    get_test_file_path
)
from ci_utils.ticket_states import (
    TicketWorkflowStep,
    TicketWorkflowStepStatus,
)


class SourcePackageUploadTestCase(unittest.TestCase):
    """Tests for `SourcePackageUpload` context manager."""

    @mock.patch('requests.post')
    def test_create_fails(self, mocked_post):
        # `create` raises `SourcePackageUploadError` in case of error.
        spu = SourcePackageUpload('1', 'http://server/')
        self.assertIsNone(spu.sandbox_url)
        mocked_post.side_effect = [
            FakeResponse(404, 'Zoing!')
        ]
        with self.assertRaises(SourcePackageUploadError) as cm:
            spu.create()
        self.assertEquals(
            'Cannot create sandbox for ticket 1 '
            '(http://server/gk/api/v1/ticket/1/sandbox/ - 404)',
            cm.exception.message)
        self.assertIsNone(spu.sandbox_url)

    @mock.patch('requests.post')
    def test_create_works(self, mocked_post):
        # `create` creates a SPU sandbox in Gatekepper storing its path.
        spu = SourcePackageUpload('1', 'http://server/')
        self.assertIsNone(spu.sandbox_url)
        mocked_post.side_effect = [
            FakeResponse(201, '/ticket/1/sandbox/UUID')
        ]
        spu.create()
        self.assertEquals(
            'http://server/gk/ticket/1/sandbox/UUID', spu.sandbox_url)

    @mock.patch('requests.put')
    def test_upload_needs_init(self, mocked_put):
        # `upload` raises an error if the sandbox context is not initialised.
        spu = SourcePackageUpload('1')
        with self.assertRaises(SourcePackageUploadError) as cm:
            spu.upload('a_path')
        self.assertEquals('SPU not initialized.', cm.exception.message)
        mocked_put.assert_not_called()

    @mock.patch('ci_cli.ticket._open_file')
    @mock.patch('requests.put')
    @mock.patch('requests.post')
    def test_upload_fail(self, mocked_post, mocked_put, mocked_of):
        # `upload` raises an error if it cannot get a tempurl from GK
        # of cannot PUT on the swift tempurl.
        mocked_post.side_effect = [
            FakeResponse(201, '/ticket/1/sandbox/UUID'),
        ]
        spu = SourcePackageUpload('1', 'http://server/')
        spu.create()
        mocked_put.side_effect = [
            FakeResponse(404, 'Not Found'),
            FakeResponse(200, 'A_TEMP_URL'),
            FakeResponse(401, 'Not Authorized'),
        ]
        with self.assertRaises(SourcePackageUploadError) as cm:
            spu.upload('not_found_filename')
        self.assertEquals(
            "Cannot get TempUrl for 'not_found_filename' "
            "(http://server/gk/ticket/1/sandbox/UUID/not_found_filename "
            "- 404)",
            cm.exception.message)
        mocked_of.return_value = 'whatever'
        with self.assertRaises(SourcePackageUploadError) as cm:
            spu.upload('a_path')
        self.assertEquals(
            "Cannot use TempUrl 'A_TEMP_URL'", cm.exception.message)

    @mock.patch('ci_cli.ticket._open_file')
    @mock.patch('requests.put')
    @mock.patch('requests.post')
    def test_upload_works(self, mocked_post, mocked_put, mocked_of):
        # `upload` returns the GK response content (created object
        # sandbox path) when it works.
        mocked_post.side_effect = [
            FakeResponse(201, '/ticket/1/sandbox/UUID'),
        ]
        spu = SourcePackageUpload('1', 'http://server/')
        spu.create()
        mocked_put.side_effect = [
            FakeResponse(200, 'A_TEMP_URL'),
            FakeResponse(201, '/a_sandbox_object_url/'),
        ]
        mocked_of.return_value = 'a_content'
        result = spu.upload('a_file')
        self.assertEquals('/a_sandbox_object_url/', result)

    @mock.patch('requests.post')
    def test_process_fail(self, mocked_post):
        # `process` raises `SourcePackageUploadError` if the sandbox is
        # not initialised.
        spu = SourcePackageUpload('1', 'http://server/')
        with self.assertRaises(SourcePackageUploadError) as cm:
            spu.process()
        self.assertEquals('SPU not initialized.', cm.exception.message)
        mocked_post.assert_not_called()

    @mock.patch('requests.post')
    def test_process_error(self, mocked_post):
        # `process` raises `SourcePackageUploadError` with the GK
        # response content (upload remote validation error).
        mocked_post.side_effect = [
            FakeResponse(201, '/ticket/1/sandbox/UUID'),
            FakeResponse(400, 'Problem!'),
        ]
        spu = SourcePackageUpload('1', 'http://server/')
        spu.create()
        with self.assertRaises(SourcePackageUploadError) as cm:
            spu.process()
        self.assertEquals(
            'SPU remote processing error: Problem!', cm.exception.message)

    @mock.patch('requests.post')
    def test_process_works(self, mocked_post):
        # `process` returns the GK response content (upload validation
        # results) when it works.
        mocked_post.side_effect = [
            FakeResponse(201, '/ticket/1/sandbox/UUID'),
            FakeResponse(201, 'Success!'),
        ]
        spu = SourcePackageUpload('1', 'http://server/')
        spu.create()
        result = spu.process()
        self.assertEquals('Success!', result)

    @mock.patch('ci_cli.ticket._open_file')
    @mock.patch('requests.put')
    @mock.patch('requests.post')
    def test_context_manager(self, mocked_post, mocked_put, mocked_of):
        # Using `SourcePackageUpload` as a context manager makes your
        # life easier!
        mocked_post.side_effect = [
            FakeResponse(201, '/ticket/1/sandbox/UUID'),
            FakeResponse(201, 'Success!'),
        ]
        mocked_put.side_effect = [
            FakeResponse(200, 'A_TEMP_URL'),
            FakeResponse(201, '/a_sandbox_object_url/'),
        ]
        mocked_of.return_value = 'a_content'

        with SourcePackageUpload('1', 'http://server/') as spu:
            result = spu.upload('a_file')

        self.assertEquals('/a_sandbox_object_url/', result)
        self.assertEquals('Success!', spu.result)


class SubTicketTestCase(TestCaseWithGnupg):
    def setUp(self):
        super(SubTicketTestCase, self).setUp()
        self.changes = get_test_file_path('foobar_0.1-1_source.changes')

    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/sourcepackage/4/')
    def test_create_sourcepackage(self, mock_post):
        new_subticket = SubTicket(owner='foobar@example.com', ticket_id=4)
        with mock.patch('sys.stdout'):
            new_subticket._parse_changes(self.changes)
        sourcepackage_uri = new_subticket._create_sourcepackage()
        mock_post.assert_called_once()
        self.assertEqual(sourcepackage_uri, '/api/v1/sourcepackage/4/')

    @mock.patch('ci_cli.utils.get_sourcepackage_uri',
                return_value='/api/v1/sourcepackage/5/')
    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/spu/38/')
    def test_create_spu(self, mock_post, mock_sp_uri):
        new_subticket = SubTicket(owner='foobar@example.com', ticket_id=3)
        with mock.patch('sys.stdout'):
            new_subticket._parse_changes(self.changes)
        new_subticket._create_spu()

    @mock.patch('ci_cli.utils.get_sourcepackage_uri', return_value='')
    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/sourcepackage/5/')
    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/spu/39/')
    def test_create_spu_sourcepackage_not_found(self, mock_spu_post, mock_post,
                                                mock_sp_uri):
        new_subticket = SubTicket(owner='foobar@example.com', ticket_id=3)
        with mock.patch('sys.stdout'):
            new_subticket._parse_changes(self.changes)
        new_subticket._create_spu()

    @mock.patch(
        'ci_cli.utils.post',
        return_value='http://www.example.com/api/v1/subticketartifact/38/',
    )
    def test_create_artifact(self, mock_post):
        new_subticket = SubTicket(owner='foobar@example.com', ticket_id=2)
        file = 'foobar_source.changes'
        location = 'http://www.example.com/path/to/foobar_source.changes'
        new_subticket._create_artifact(file, location)
        mock_post.assert_called_once()

    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/subticket/1/')
    def test_create_subticket(self, mock_post):
        new_subticket = SubTicket(owner='foobar@example.com', ticket_id=1)
        new_subticket._create_subticket()


class TicketTestCase(MainScriptTestCase):

    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/ticket/38/')
    def test_create_ticket(self, mock_post):
        changes = get_test_file_path('foobar_0.1-1_source.changes')
        args = ['create_ticket', '-t', 'New feature', '-b', '1234', '-o',
                'someone@example.com', '-d', 'This is a cool new feature',
                '-s', changes]

        new_ticket_ = Ticket()
        args = self.cli.parse_arguments(args)
        new_ticket_._create_ticket(args)

    @mock.patch('ci_cli.ticket.SubTicket._process')
    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/ticket/38/')
    @mock.patch('ci_cli.utils.patch')
    def test_add_new_ticket(self, mock_patch, mock_post, mock_process):
        """Verify cli create_ticket command for one .changes."""
        changes = get_test_file_path('foobar_0.1-1_source.changes')
        args = [
            'create_ticket',
            '-t', 'New feature',
            '-b', '1234',
            '-o', 'someone@example.com',
            '-d', 'This is a cool new feature',
            '-s', changes,
        ]

        args = self.cli.parse_arguments(args)
        with mock.patch('sys.stdout'):
            new_ticket_ = new_ticket(args)
        self.assertEquals(mock_process.call_count, 1)
        self.assertEquals(mock_post.call_count, 1)
        self.assertEquals(mock_patch.call_count, 1)
        self.assertEquals(len(new_ticket_.subtickets), 1)
        self.assertTrue("foobar" in new_ticket_.subtickets)

    @mock.patch('ci_cli.ticket.SubTicket._process')
    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/ticket/38/')
    @mock.patch('ci_cli.utils.patch')
    def test_add_new_ticket_no_queueing(self, mock_patch, mock_post,
                                        mock_process):
        # Users may choose to not start a ticket by passing a '-w'
        # options on creation.
        changes = get_test_file_path('foobar_0.1-1_source.changes')
        args = [
            'create_ticket',
            '-t', 'New feature',
            '-b', '1234',
            '-o', 'someone@example.com',
            '-d', 'This is a cool new feature',
            '-s', changes,
            '-w'
        ]

        args = self.cli.parse_arguments(args)
        # Use a global variable to store the newly created ticket
        # and simultaneously capture the creation messages.
        global _ticket
        _ticket = None

        def _do_create():
            global _ticket
            _ticket = new_ticket(args)

        with capture_stdout(_do_create) as cm:
            # User is instructed how to start the NEW ticket later.
            self.assertEqual(
                ['You have successfully created a ticket on Ubuntu CI '
                 'Engine under the ID #38, but it has not started yet.',
                 'Once you are ready, you can submit it for processing '
                 'by calling:',
                 '',
                 '\t{} --url None update_ticket -i 38 --start'.format(
                     sys.argv[0]),
                 ''],
                cm.splitlines()[-5:])
        # Ticket was processed and created, but not queued.
        self.assertEquals(1, len(_ticket.subtickets))
        self.assertTrue("foobar" in _ticket.subtickets)
        mock_post.assert_called_once_with(
            '{}/api/v1/ticket/'.format(utils.CI_URL),
            data={
                'owner': 'someone@example.com',
                'series': 'trusty',
                'bug_id': '1234',
                'description': 'This is a cool new feature',
                'title': 'New feature',
            }
        )
        mock_patch.assert_not_called()

    @mock.patch('ci_cli.ticket.SubTicket._process')
    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/ticket/39/')
    @mock.patch('ci_cli.utils.patch')
    def test_add_new_ticket_multiple_subtickets(self, mock_patch, mock_post,
                                                mock_process):
        """Verify cli create_ticket command for multiple .changes.

        This also tests the case two sources are targeted to the same series.
        """
        changes_1 = get_test_file_path('foobar_0.1-1_source.changes')
        changes_2 = get_test_file_path('barfoo_0.1-1_source.changes')

        args = [
            'create_ticket',
            '-t', 'New feature',
            '-b', '4321',
            '-o', 'someone@example.com',
            '-d', 'This is a cool new feature',
            '-s', changes_1,
            '-s', changes_2,
        ]

        args = self.cli.parse_arguments(args)
        with mock.patch('sys.stdout'):
            new_ticket_ = new_ticket(args)
        self.assertEquals(mock_process.call_count, 2)
        self.assertEquals(mock_post.call_count, 1)
        self.assertEquals(mock_patch.call_count, 1)

        self.assertEquals(len(new_ticket_.subtickets), 2)
        self.assertTrue("foobar" in new_ticket_.subtickets)
        self.assertTrue("barfoo" in new_ticket_.subtickets)
        self.assertEquals(u'trusty', new_ticket_.suite)


class StableMainScriptTestCase(MainScriptTestCase):

    def setUp(self):
        """Lazily import 'utils' and set CI_URL for tests.

        This is necessary because the CLI is executed inside the test suite
        and sets CI_URL. So depending on the test run order it might or
        might not be set.
        """
        super(StableMainScriptTestCase, self).setUp()
        import ci_cli.utils
        ci_cli.utils.CI_URL = 'http://ci-airline.ubuntu.com'


class TicketUpdateTestCase(StableMainScriptTestCase):

    def test_update_empty(self):
        # Ticket update attempts with no 'updates' do not reach the network.
        args_list = ['update_ticket', '-i', '1']
        args = self.cli.parse_arguments(args_list)
        update_ticket(args)

    @mock.patch('ci_cli.utils.get')
    @mock.patch('ci_cli.utils.patch')
    def test_update_only_given_fields(self, mocked_patch, mocked_get):
        # Updating attempts only consider given fields for PATCH.
        mocked_get.return_value = {
            'current_workflow_step': TicketWorkflowStep.NEW,
        }
        args_list = [
            'update_ticket',
            '-i', '1',
            '-t', 'Updated!',
        ]
        args = self.cli.parse_arguments(args_list)
        with mock.patch('sys.stdout'):
            update_ticket(args)

        mocked_get.assert_called_once_with(
            'http://ci-airline.ubuntu.com/api/v1/fullticket/1/')
        mocked_patch.assert_called_once_with(
            'http://ci-airline.ubuntu.com/api/v1/ticket/1/',
            data={'title': 'Updated!'})

    @mock.patch('ci_cli.utils.get')
    @mock.patch('ci_cli.utils.patch')
    def test_update_start(self, mocked_patch, mocked_get):
        # A NEW ticket can be started using the CLI.
        mocked_get.return_value = {
            'current_workflow_step': str(TicketWorkflowStep.NEW),
        }
        args_list = [
            'update_ticket',
            '-i', '1',
            '-t', 'Updated!',
            '-y'
        ]
        args = self.cli.parse_arguments(args_list)
        with mock.patch('sys.stdout'):
            update_ticket(args)

        mocked_get.assert_called_once_with(
            'http://ci-airline.ubuntu.com/api/v1/fullticket/1/')
        mocked_patch.assert_called_once_with(
            'http://ci-airline.ubuntu.com/api/v1/ticket/1/',
            data={'current_workflow_step': TicketWorkflowStep.QUEUED.value,
                  'status': TicketWorkflowStepStatus.QUEUED.value,
                  'title': 'Updated!'})

    @mock.patch('ci_cli.utils.get')
    @mock.patch('ci_cli.utils.patch')
    def test_update_can_only_start_new(self, mocked_patch, mocked_get):
        # Only NEW tickets can be started.
        mocked_get.return_value = {
            'current_workflow_step': str(TicketWorkflowStep.QUEUED),
            'status': str(TicketWorkflowStepStatus.QUEUED),
        }
        args_list = [
            'update_ticket',
            '-i', '1',
            '-t', 'Updated!',
            '-y'
        ]
        args = self.cli.parse_arguments(args_list)
        with self.assertRaises(QueueingTicketError) as cm:
            update_ticket(args)
        self.assertEqual(
            'Ticked #1 has already started (Queued - Not started) and '
            'cannot be re-started manually.',
            str(cm.exception))
        mocked_get.assert_called_once_with(
            'http://ci-airline.ubuntu.com/api/v1/fullticket/1/')
        mocked_patch.assert_not_called()

    @mock.patch('ci_cli.utils.local_raw_input', return_value='')
    @mock.patch('ci_cli.ticket.SubTicket._process')
    @mock.patch('ci_cli.utils.get')
    @mock.patch('ci_cli.utils.patch')
    def test_update_sources(self, mocked_patch, mocked_get, mocked_process,
                            mocked_ri):
        # New sources can be added to a ticket and requires user
        # confirmation (defaults to yes).
        mocked_get.return_value = {
            'series': 'trusty',
        }
        foo_changes = get_test_file_path('foobar_0.1-1_source.changes')
        bar_changes = get_test_file_path('barfoo_0.1-1_source.changes')
        args_list = [
            'update_ticket',
            '-i', '1',
            '-t', 'Updated!',
            '-s', foo_changes,
            '-s', bar_changes,
        ]
        args = self.cli.parse_arguments(args_list)

        global ticket
        ticket = None

        def _do_update():
            global ticket
            ticket = update_ticket(args)

        with capture_stdout(_do_update) as cm:
            # User is informed about the new sources.
            self.assertEqual(
                ['You are about to include the following new source(s) '
                 'to the ticket #1 (Updated!):',
                 '  * foobar_0.1-1',
                 '  * barfoo_0.1-1',
                 'Ticket #1 updated.'],
                cm.splitlines()[-4:])

        # Both sources were parsed (as subtickets) and processed.
        self.assertEqual(
            ['barfoo', 'foobar'], sorted(ticket.subtickets.keys()))
        self.assertEqual('trusty', ticket.suite)
        mocked_process.assert_called_once()
        # Ticket was updated after adding new sources.
        mocked_patch.assert_called_once_with(
            'http://ci-airline.ubuntu.com/api/v1/ticket/1/',
            data={'title': 'Updated!'})

    @mock.patch('ci_cli.utils.local_raw_input', return_value='n')
    @mock.patch('ci_cli.ticket.SubTicket._process')
    @mock.patch('ci_cli.utils.get')
    @mock.patch('ci_cli.utils.patch')
    def test_update_sources_not_confirmed(self, mocked_patch, mocked_get,
                                          mocked_process, mocked_ri):
        # If not confirmed, the update including new sources is not executed.
        mocked_get.return_value = {
            'series': 'trusty',
        }
        foo_changes = get_test_file_path('foobar_0.1-1_source.changes')
        args_list = [
            'update_ticket',
            '-i', '1',
            '-t', 'Not confirmed',
            '-s', foo_changes,
        ]
        args = self.cli.parse_arguments(args_list)
        with LogCapture() as lc:
            update_ticket(args)
        last_log = lc.records[-1]
        self.assertEqual('WARNING', last_log.levelname)
        self.assertEqual('Update dismissed.', last_log.msg)
        mocked_process.assert_not_called()
        mocked_patch.assert_not_called()

    @mock.patch('ci_cli.utils.get')
    @mock.patch('ci_cli.utils.patch')
    def test_update_sources_different_suite(self, mocked_patch, mocked_get):
        # New sources must be coherent with the ticket series.
        mocked_get.return_value = {
            'series': 'precise',
        }
        foo_changes = get_test_file_path('foobar_0.1-1_source.changes')
        args_list = [
            'update_ticket',
            '-i', '1',
            '-t', 'Updated!',
            '-s', foo_changes,
        ]
        args = self.cli.parse_arguments(args_list)
        with self.assertRaises(MismatchSuitesError) as cm:
            update_ticket(args)
        self.assertEqual(
            'Submitted sources are not targeted to ticket suite (precise)',
            str(cm.exception))
        mocked_patch.assert_not_called()

    @mock.patch('ci_cli.utils.get')
    @mock.patch('ci_cli.utils.patch')
    def test_update_sources_uniqueness(self, mocked_patch, mocked_get):
        # New sources must be unique.
        mocked_get.return_value = {
            'series': 'trusty',
            'subticket': [{
                'source_package_upload': {
                    'sourcepackage': {'name': 'foobar'},
                    'version': '0.1-1',
                }
            }]
        }
        foo_changes = get_test_file_path('foobar_0.1-1_source.changes')
        args_list = [
            'update_ticket',
            '-i', '1',
            '-t', 'Updated!',
            '-s', foo_changes,
        ]
        args = self.cli.parse_arguments(args_list)
        with self.assertRaises(SourceAlreadyUploadedError) as cm:
            update_ticket(args)
        self.assertEqual(
            'Submitted source foobar_0.1-1 was already uploaded '
            'to ticket #1',
            str(cm.exception))
        mocked_patch.assert_not_called()


class TicketStatusTestCase(StableMainScriptTestCase):

    ticket_data = {
        "current_workflow_step": str(TicketWorkflowStep.IMAGE_BUILDING),
        "status": str(TicketWorkflowStepStatus.INPROGRESS),
        "title": "My cool feature",
        "id": '100',
        "description": "my new feature",
        "owner": "user@example.com",
        "created": "Wed, 12 Mar 2014 02:18:28 +0000",
        "updated": "Wed, 12 Mar 2014 02:30:22 +0000",
        "bug_id": "12345",
        "uuid": "f7eba912-31d0-11e4-a2da-fa163eac7247",
    }

    tickets_data = {
        "meta": {'total_count': 4},
        "objects": [{
            "id": 5,
            "current_workflow_step": str(TicketWorkflowStep.IMAGE_BUILDING),
            "status": str(TicketWorkflowStepStatus.INPROGRESS),
            "title": "Cool feature",
            "uuid": "f7eba912-31d0-11e4-a2da-fa163eac7247",
        }, {
            "id": 6,
            "current_workflow_step": str(TicketWorkflowStep.IMAGE_BUILDING),
            "status": str(TicketWorkflowStepStatus.INPROGRESS),
            "title": "Bug fix",
            "uuid": "bbeba912-31d0-11e4-a2da-fa163eac7247",
        }, {
            "id": 7,
            "current_workflow_step": str(TicketWorkflowStep.IMAGE_BUILDING),
            "status": str(TicketWorkflowStepStatus.INPROGRESS),
            "title": "Upstream fix",
            "uuid": "cceba912-31d0-11e4-a2da-fa163eac7247",
        }, {
            "id": 9,
            "current_workflow_step": str(TicketWorkflowStep.IMAGE_BUILDING),
            "status": str(TicketWorkflowStepStatus.INPROGRESS),
            "title": "Other feature",
            "uuid": "ddeba912-31d0-11e4-a2da-fa163eac7247",
        }]
    }

    @mock.patch('requests.get')
    def test_get_single_ticket_status(self, mock_get):
        # Users are able to print single ticket details.
        mock_get.side_effect = [
            FakeResponse(200, json.dumps(self.ticket_data))
        ]
        args = self.cli.parse_arguments(
            ['status', '-i', 'f7eba912-31d0-11e4-a2da-fa163eac7247'])
        with capture_stdout(args.func, args) as cm:
            self.assertEqual(
                ['Ticket f7eba912-31d0-11e4-a2da-fa163eac7247 - '
                 'My cool feature',
                 'Current landing step: Image building - In progress',
                 'Description: my new feature',
                 'Owner: user@example.com',
                 'Created: Wed, 12 Mar 2014 02:18:28 +0000',
                 'Updated: Wed, 12 Mar 2014 02:30:22 +0000',
                 'Bug #: 12345'],
                str(cm).splitlines())
        mock_get.assert_called_once_with(
            'http://ci-airline.ubuntu.com/api/v1/ticket/'
            'f7eba912-31d0-11e4-a2da-fa163eac7247/')

    @mock.patch('requests.get')
    def test_get_ticket_status_unicode(self, mock_get):
        # Ticket details are printed as 'UTF-8'.
        data = self.ticket_data.copy()
        data.update({
            'title': 'Testing unicode in title: joão',
        })
        mock_get.side_effect = [
            FakeResponse(200, json.dumps(data))
        ]
        args = self.cli.parse_arguments(
            ['status', '-i', 'f7eba912-31d0-11e4-a2da-fa163eac7247'])
        with capture_stdout(args.func, args) as cm:
            self.assertEqual(
                ['Ticket f7eba912-31d0-11e4-a2da-fa163eac7247 - '
                 'Testing unicode in title: '
                 'joão'.encode('utf-8'),
                 'Current landing step: Image building - In progress',
                 'Description: my new feature',
                 'Owner: user@example.com',
                 'Created: Wed, 12 Mar 2014 02:18:28 +0000',
                 'Updated: Wed, 12 Mar 2014 02:30:22 +0000',
                 'Bug #: 12345'],
                str(cm).splitlines())
        mock_get.assert_called_once_with(
            'http://ci-airline.ubuntu.com/api/v1/ticket/'
            'f7eba912-31d0-11e4-a2da-fa163eac7247/')

    @mock.patch('requests.get')
    def test_get_single_failed_ticket_status(self, mock_get):
        # Failed tickets have a more concise status line (only the
        # ticket status).
        data = self.ticket_data.copy()
        data.update({
            "current_workflow_step": str(TicketWorkflowStep.FAILED),
            "status": str(TicketWorkflowStepStatus.FAILED),
        })
        mock_get.side_effect = [
            FakeResponse(200, json.dumps(data))
        ]
        args = self.cli.parse_arguments(
            ['status', '-i', 'f7eba912-31d0-11e4-a2da-fa163eac7247'])
        with capture_stdout(args.func, args) as cm:
            self.assertEqual(
                ['Ticket f7eba912-31d0-11e4-a2da-fa163eac7247 - '
                 'My cool feature',
                 'Failed',
                 'Description: my new feature',
                 'Owner: user@example.com',
                 'Created: Wed, 12 Mar 2014 02:18:28 +0000',
                 'Updated: Wed, 12 Mar 2014 02:30:22 +0000',
                 'Bug #: 12345'],
                str(cm).splitlines())
        mock_get.assert_called_once()

    @mock.patch('requests.get')
    def test_get_all_tickets_status(self, mock_get):
        # Users can list the last batch of tickets by not setting any
        # limit/offset arguments.
        mock_get.side_effect = [
            FakeResponse(200, json.dumps(self.tickets_data)),
            FakeResponse(200, json.dumps(self.tickets_data)),
        ]
        args = self.cli.parse_arguments(['status'])
        with capture_stdout(args.func, args) as cm:
            self.assertEqual(
                ["Ticket f7eba912-31d0-11e4-a2da-fa163eac7247: "
                 "Cool feature is Image building (In progress)",
                 "Ticket bbeba912-31d0-11e4-a2da-fa163eac7247: "
                 "Bug fix is Image building (In progress)",
                 "Ticket cceba912-31d0-11e4-a2da-fa163eac7247: "
                 "Upstream fix is Image building (In progress)",
                 "Ticket ddeba912-31d0-11e4-a2da-fa163eac7247: "
                 "Other feature is Image building (In progress)"],
                str(cm).splitlines())
        # The first request is issued for calculating the offset for the
        # last batch, then the second one is issued for retrieving it.
        self.assertEqual(
            [mock.call('http://ci-airline.ubuntu.com/api/v1/ticket/'
                       '?limit=1&offset=0'),
             mock.call('http://ci-airline.ubuntu.com/api/v1/ticket/'
                       '?limit=20&offset=0')],
            mock_get.call_args_list)

    @mock.patch('requests.get')
    def test_get_all_tickets_limit_and_offset(self, mock_get):
        # Users can list tickets with limit and offset.
        mock_get.side_effect = [
            FakeResponse(200, json.dumps(self.tickets_data)),
        ]
        args = self.cli.parse_arguments(['status', '-l', '2', '-o', '2'])
        args.func(args)
        # It simply relies on the TS(tastypie) pagination.
        mock_get.assert_called_once_with(
            'http://ci-airline.ubuntu.com/api/v1/ticket/?limit=2&offset=2')
