#!/usr/bin/env python
# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.

# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import mock
import unittest

from ci_cli import utils
from ci_cli.tests import FakeResponse
from requests import exceptions


config_contents = """auth_url: http://example.com
auth_user: user
auth_password: pass
auth_tenant_name: user_project
auth_region: region1
ci_url: http://example.com/ci
"""


class TestBasicRequest(unittest.TestCase):

    @mock.patch('requests.get')
    def test_get(self, mocked_get):
        mocked_get.side_effect = [FakeResponse(200, '{"foo": ["bar", "baz"]}')]
        result = utils.get("foo")
        self.assertEquals(result, {u'foo': [u'bar', u'baz']})
        mocked_get.assert_called_once()

    @mock.patch('requests.get')
    def test_get_fails(self, mocked_get):
        mocked_get.side_effect = [FakeResponse(502, {}, "Failed get()!")]
        with self.assertRaises(exceptions.HTTPError) as cm:
            utils.get("foo")
        mocked_get.assert_called_once()
        self.assertEquals(
            cm.exception.message, "502 Server Error: Failed get()!")

    @mock.patch('requests.get')
    def test_get_returns_invalid_content(self, mocked_get):
        mocked_get.side_effect = [FakeResponse(200, "Ohh, this is invalid")]
        with self.assertRaises(ValueError) as cm:
            utils.get("foo")
        mocked_get.assert_called_once()
        self.assertEquals(
            cm.exception.message, "No JSON object could be decoded")

    @mock.patch('requests.patch')
    def test_patch(self, mocked_patch):
        utils.patch("foo", {}, {})
        mocked_patch.assert_called_once()

    @mock.patch('requests.patch')
    def test_patch_fails(self, mocked_patch):
        mocked_patch.side_effect = [FakeResponse(502, {}, "Failed patch()!")]
        with self.assertRaises(exceptions.HTTPError) as cm:
            utils.patch("foo", {}, {})
        mocked_patch.assert_called_once()
        self.assertEquals(
            cm.exception.message, "502 Server Error: Failed patch()!")

    @mock.patch('requests.post')
    def test_post(self, mocked_post):
        utils.post("foo", {}, {})
        mocked_post.assert_called_once()

    @mock.patch('requests.post')
    def test_post_fails(self, mocked_post):
        mocked_post.side_effect = [FakeResponse(502, {}, "Failed post()!")]
        with self.assertRaises(exceptions.HTTPError) as cm:
            utils.post("foo", {}, {})
        mocked_post.assert_called_once()
        self.assertEquals(
            cm.exception.message, "502 Server Error: Failed post()!")


class UtilsTestCase(unittest.TestCase):

    def test_parse_id(self):
        location = 'http://www.example.com/api/v1/ticket/5/'
        id = utils.parse_id(location)
        self.assertEqual(id, '5')

    def test_parse_id_invalid_response(self):
        location = 'http://www.example.com/api/v1/ticket/'
        self.assertRaises(ValueError, utils.parse_id, location)

    def test_parse_id_not_int(self):
        location = 'http://www.example.com/api/v1/ticket/abc/123/'
        self.assertRaises(ValueError, utils.parse_id, location)

    @mock.patch('requests.get')
    def test_get_sourcepackage_uri(self, mock_requests):
        data = {"objects": [{"resource_uri": "/api/v1/sourcepackage/4/"}]}
        mock_requests.side_effect = [
            FakeResponse(200, json.dumps(data))
        ]
        uri = utils.get_sourcepackage_uri('foobar')
        mock_requests.assert_called_once()
        self.assertEquals(uri, '/api/v1/sourcepackage/4/')

    @mock.patch('requests.get')
    def test_get_sourcepackage_uri_not_found(self, mock_requests):
        data = {"objects": []}
        mock_requests.side_effect = [
            FakeResponse(200, json.dumps(data))
        ]
        uri = utils.get_sourcepackage_uri('foobar')
        mock_requests.assert_called_once()
        self.assertEquals(uri, '')

    def test_assert_valid_package_list_single(self):
        utils.assert_valid_package_list('foo')

    def test_assert_valid_package_list_three(self):
        utils.assert_valid_package_list('foo,bar,baz')

    def test_assert_valid_package_list_empty(self):
        utils.assert_valid_package_list(package_list_str=None)

    def test_assert_valid_package_list_with_space(self):
        with self.assertRaises(utils.InputError):
            utils.assert_valid_package_list('bar, baz')


class UserConfirmationTestCase(unittest.TestCase):

    @mock.patch('ci_cli.utils.local_raw_input')
    def test_defaults(self, mocked_raw):
        # `get_user_confirmation` provides sensible defaults.
        mocked_raw.return_value = 'y'
        self.assertTrue(utils.get_user_confirmation())
        mocked_raw.assert_called_once_with('Confirm ? [y/n]: ')

    @mock.patch('ci_cli.utils.local_raw_input')
    def test_not_confirmed(self, mocked_raw):
        # `get_user_confirmation` accepts negation.
        mocked_raw.return_value = 'n'
        self.assertFalse(utils.get_user_confirmation())
        mocked_raw.assert_called_once_with('Confirm ? [y/n]: ')

    @mock.patch('ci_cli.utils.local_raw_input')
    def test_nags_until_valid(self, mocked_raw):
        # `get_user_confirmation` keep asking until it gets a valid answer.
        mocked_raw.side_effect = ['wot', 'nahh', 'ok', 'n']
        self.assertFalse(utils.get_user_confirmation())
        self.assertEqual(4, mocked_raw.call_count)

    @mock.patch('ci_cli.utils.local_raw_input')
    def test_default_answer(self, mocked_raw):
        # `get_user_confirmation` allows default answer and hints to it
        # in the confirmation message.
        mocked_raw.return_value = ''
        self.assertTrue(
            utils.get_user_confirmation(default='y'))
        mocked_raw.assert_called_once_with('Confirm ? [Y/n]: ')

    @mock.patch('ci_cli.utils.local_raw_input')
    def test_custom_message(self, mocked_raw):
        # `get_user_confirmation` allows custom messages.
        mocked_raw.return_value = 'y'
        self.assertTrue(
            utils.get_user_confirmation('Really ?'))
        mocked_raw.assert_called_once_with('Really ? [y/n]: ')

    @mock.patch('ci_cli.utils.local_raw_input')
    def test_custom_validation(self, mocked_raw):
        # `get_user_confirmation` allows custom answer validation.
        mocked_raw.return_value = 'sure'
        self.assertTrue(
            utils.get_user_confirmation(
                valid=['sure', 'no-way'], confirmed=['sure']))
        mocked_raw.assert_called_once_with('Confirm ? [sure/no-way]: ')

    def test_valid_must_be_a_list(self):
        # `get_user_confirmation` valid options must be a list.
        with self.assertRaises(ValueError) as cm:
            utils.get_user_confirmation(valid='y')
        self.assertEqual(
            "'valid' must be a list of options.", str(cm.exception))

    def test_confirmed_must_be_valid(self):
        # `get_user_confirmation` confirmed options must be a list of
        # 'valid' options.
        with self.assertRaises(ValueError) as cm:
            utils.get_user_confirmation(confirmed='y')
        self.assertEqual(
            "'confirmed' must be a list of options.", str(cm.exception))
        with self.assertRaises(ValueError) as cm:
            utils.get_user_confirmation(confirmed=['ok'])
        self.assertEqual(
            "'confirmed' options (['ok']) not in 'valid' (['y', 'n']).",
            str(cm.exception))

    def test_default_must_be_valid(self):
        # `get_user_confirmation` provided default value should be valid.
        with self.assertRaises(ValueError) as cm:
            utils.get_user_confirmation(default='nah')
        self.assertEqual(
            "'default' option (nah) not in 'valid' (['y', 'n']).",
            str(cm.exception))
