# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import logging
import os
import requests
import sys
import urlparse

from ci_cli import utils
from ci_utils.file_handler import ChangesProcessor
from ci_utils.ticket_states import (
    TicketWorkflowStep,
    TicketWorkflowStepStatus,
)


log = logging.getLogger('ubuntu-ci')


def _open_file(filepath):
    """Local file opener for making tests easier."""
    return open(filepath)


class SourceAlreadyUploadedError(Exception):
    """Raised when submitted source is already part of the ticket."""


class MismatchSuitesError(Exception):
    """Raised when submitted packages target multiple series."""


class UnsupportedSeriesError(Exception):
    """Raised when package series isn't supported."""


class UnsupportedSuiteError(Exception):
    """Raised when package series isn't supported."""


class SourcePackageUploadError(Exception):
    """Raised where there is an error during the SPU."""


class QueueingTicketError(Exception):
    """Raised when ticket failed to be set to Queued."""


class SourcePackageUpload(object):
    """Context manager helper for performing SPUs via Gatekeeper."""

    def __init__(self, ticket_id, ci_url=None):
        if ci_url is None:
            ci_url = utils.CI_URL
        self.gk_url = urlparse.urljoin(ci_url, '/gk')
        self.ticket_id = ticket_id
        self.sandbox_url = None
        self.result = None

    def __enter__(self):
        """Creates a remote sandbox context."""
        self.create()
        return self

    def __exit__(self, exc, value, tb):
        """Process the context sandbox."""
        self.result = self.process()

    def create(self):
        """Request a Gatekeeper Sandbox."""
        create_url = self.gk_url + '/api/v1/ticket/{}/sandbox/'.format(
            self.ticket_id)
        response = requests.post(create_url)
        if response.status_code != 201:
            raise SourcePackageUploadError(
                'Cannot create sandbox for ticket {} ({} - {})'.format(
                    self.ticket_id, create_url, response.status_code))
        sandbox_path = response.content
        self.sandbox_url = self.gk_url + sandbox_path

    def process(self):
        """Process the initialised Sandbox."""
        if self.sandbox_url is None:
            raise SourcePackageUploadError('SPU not initialized.')
        response = requests.post(self.sandbox_url)
        if response.status_code != 201:
            raise SourcePackageUploadError(
                'SPU remote processing error: {}'.format(response.content))
        return response.content

    def upload(self, filepath):
        """Upload the given file to the context Sandbox."""
        if self.sandbox_url is None:
            raise SourcePackageUploadError('SPU not initialized.')
        dest_url = self.sandbox_url + '/' + os.path.basename(filepath)
        response = requests.put(dest_url)
        if response.status_code != 200:
            filename = os.path.basename(filepath)
            raise SourcePackageUploadError(
                'Cannot get TempUrl for \'{}\' ({} - {})'.format(
                    filename, dest_url, response.status_code))
        temp_url = response.content
        response = requests.put(temp_url, data=_open_file(filepath))
        if response.status_code != 201:
            raise SourcePackageUploadError(
                'Cannot use TempUrl \'{}\''.format(temp_url))
        return response.content


def upload_files(ticket_id, files):
    """Upload files to a ticket container via Gatekeeper."""
    with SourcePackageUpload(ticket_id) as spu:
        for file_path in files:
            filename = os.path.basename(file_path)
            log.info('Uploading {} ...'.format(filename))
            try:
                spu.upload(file_path)
            except SourcePackageUploadError as exc:
                log.error(exc.message)
                raise
    ticket_url = urlparse.urljoin(utils.CI_URL, spu.result)
    log.info('Done! Access {} for more information.'.format(ticket_url))


class SubTicket():

    def __init__(self, owner, ticket_id=None):
        self.owner = owner
        self.ticket_id = ticket_id
        self.sourcepackage = ''
        self.files = ''
        self.version = ''
        self.architecture = ''

        self.spu_uri = ''
        self.subticket_uri = ''

    @property
    def ticket_uri(self):
        if self.ticket_id:
            return utils.TICKET_BASE + str(self.ticket_id) + '/'
        return ''

    def _parse_changes(self, changes_filepath):
        log.info("Parsing {}...".format(os.path.basename(changes_filepath)))
        changes = ChangesProcessor(
            changes_filepath=changes_filepath)
        log.info("Validating .changes...")
        self.files = changes.process()
        self.version = changes.source_version
        self.suite = changes.suite
        self.series = self.suite.split("-")[0]
        self.sourcepackage = changes.source_package_name
        self.version = changes.source_version
        self.architecture = changes.architecture
        log.info("{}_{} parsed and validated.".format(
            self.sourcepackage, self.version))

    def _process(self):
        self._create_spu()
        self._create_subticket()
        self._create_artifacts()

    def _create_artifacts(self):
        """Create artifacts for this ticket."""
        upload_files(self.ticket_id, self.files)
        for file_path in self.files:
            filename = os.path.basename(file_path)
            location = os.path.join(
                'ticket-{}'.format(self.ticket_id), filename)
            self._create_artifact(filename, location)

    def _create_sourcepackage(self):
        data = {
            "name": self.sourcepackage,
        }
        location = utils.post(
            utils.CI_URL + utils.SOURCEPACKAGE_BASE, data=data)
        sourcepackage_id = utils.parse_id(location)
        return utils.SOURCEPACKAGE_BASE + sourcepackage_id + '/'

    def _create_spu(self):
        sourcepackage_uri = utils.get_sourcepackage_uri(self.sourcepackage)
        if sourcepackage_uri == '':
            sourcepackage_uri = self._create_sourcepackage()
        data = {
            "sourcepackage": sourcepackage_uri,
            "version": self.version,
            "architecture": self.architecture,
        }
        location = utils.post(utils.CI_URL + utils.SPU_BASE, data=data)
        self.spu_id = utils.parse_id(location)
        self.spu_uri = utils.SPU_BASE + self.spu_id + '/'
        log.info("Created source package upload: %s" % location)
        return location

    def _create_subticket(self):
        data = {
            "ticket": self.ticket_uri,
            "assignee": self.owner,
            "source_package_upload": self.spu_uri,
        }
        location = utils.post(utils.CI_URL + utils.SUBTICKET_BASE, data=data)
        self.subticket_id = utils.parse_id(location)
        self.subticket_uri = utils.SUBTICKET_BASE + self.subticket_id + '/'
        log.info("Created subticket: %s" % location)
        return location

    def _create_artifact(self, file, location):
        data = {
            "type": "SPU",
            "reference": location,
            "name": file,
            "subticket": self.subticket_uri,
        }
        location_ = utils.post(utils.CI_URL + utils.SUBTICKET_ARTIFACT_BASE,
                               data=data)
        log.info("Created artifact: %s" % location_)


class Ticket():

    def __init__(self):
        self.ticket_id = ''
        self.ticket_uri = ''
        self.uploads = []
        self.subtickets = {}
        self.suite = ''

    def _create_ticket(self, args):
        data = {
            "title": args.title,
            "owner": args.owner,
            "description": args.description,
            "series": self.suite.split("-")[0],
            "bug_id": args.bug,
        }
        location = utils.post(utils.CI_URL + utils.TICKET_BASE, data=data)
        self.ticket_id = location.split('/')[-2]
        self.ticket_uri = utils.TICKET_BASE + self.ticket_id + '/'

    def _queue_ticket(self):
        """Queues the context ticket for processing.

        After a ticket has been created, it needs to be updated to have a
        status of QUEUED in order to be processed.
        """
        data = {
            "current_workflow_step": TicketWorkflowStep.QUEUED.value,
            "status": TicketWorkflowStepStatus.QUEUED.value,
        }
        urlbase = utils.CI_URL + utils.TICKET_BASE + self.ticket_id + '/'
        try:
            utils.patch(urlbase, data=data)
        except Exception as exc:
            raise QueueingTicketError(
                "Ticket was created successfully but there was a problem "
                "adding it to the queue: {}. This should be a transient "
                "error, please contact an admin. Your ticket number is "
                "{}.".format(exc, self.ticket_id))

        print("You have successfully submitted a ticket to the Ubuntu CI "
              "Engine. Your ticket id is %s." % self.ticket_id)

    def _parse_sources(self, args):
        """Parse given sources into subtickets.

        Raises `MismatchSuitesError` if submitted sources are not targeted
        to the same suite.
        """
        suites = set()
        # First, process all sources and validate the series.
        for source in args.sources:
            subticket = SubTicket(owner=args.owner)
            subticket._parse_changes(source)
            suites.add(subticket.suite)
            self.subtickets[subticket.sourcepackage] = subticket
        if len(suites) > 1:
            raise MismatchSuitesError(
                "All submitted packages have to target the same suite. "
                "You submitted packages with suites: {}".format(
                    ", ".join(suites)))
        self.suite = suites.pop()

    def _process_subtickets(self):
        """ """
        for subticket in self.subtickets.itervalues():
            subticket.ticket_id = self.ticket_id
            subticket._process()

    def _calculate_updates(self, args):
        """Returns a dictionary with the tickets attributes to be updated."""
        payload = {
            'title': args.title,
            'description': args.description,
            'bug_id': args.bug,
            'owner': args.owner,
        }

        # Add the specific options for starting a ticket if requested.
        if args.start:
            payload.update({
                "current_workflow_step": TicketWorkflowStep.QUEUED.value,
                "status": TicketWorkflowStepStatus.QUEUED.value,
            })

        # Remove absent values, so the corresponding ticket values are
        # preserved.
        updates = payload.copy()
        for k, v in payload.iteritems():
            if v is None:
                del updates[k]

        return updates

    def _get_tickets(self, full=False, limit=20, offset=0):
        """Returns the representation of a list of tickets."""
        # Zero means no limit for tastypie pagination, not allowed.
        if limit <= 0:
            limit = 20
        # Negative offsets are casted to ZERO.
        if offset < 0:
            offset = 0
        if full:
            model = 'fullticket'
        else:
            model = 'ticket'
        path = os.path.join(utils.API_URL, model, '')
        path = path + '?limit={}&offset={}'.format(limit, offset)
        url = urlparse.urljoin(utils.CI_URL, path)
        return utils.get(url)

    def _get_ticket(self, full=False):
        """Returns the representation of the context ticket.

        :params full: whether to use the 'fullticket' or the 'ticket'
            model. Defaults False (ticket).
        """
        if full:
            model = 'fullticket'
        else:
            model = 'ticket'
        path = os.path.join(utils.API_URL, model, self.ticket_id, '')
        url = urlparse.urljoin(utils.CI_URL, path)
        return utils.get(url)

    def create(self, args):
        """Creates a new ticket from the given command-line arguments."""
        # Parses and validates given sources.
        self._parse_sources(args)

        # Create the new ticket.
        self._create_ticket(args)

        # Creates subtickets for the processed sources.
        self._process_subtickets()

        # Queue the new ticket for processing, if not requested otherwise.
        if args.wait:
            print('You have successfully created a ticket on Ubuntu CI '
                  'Engine under the ID #{}, but it has not started '
                  'yet.'.format(self.ticket_id))
            print ('Once you are ready, you can submit it for processing '
                   'by calling:\n\n'
                   '\t{} --url {} update_ticket -i {} --start\n'.format(
                       sys.argv[0], args.url, self.ticket_id))
            return
        self._queue_ticket()

    def update(self, args):
        """Updates an existing ticket from the given arguments."""
        self.ticket_id = args.id

        # Calculate updates according to the given args. Dismiss if
        # there is nothing to update.
        updates = self._calculate_updates(args)
        if not updates and not args.sources:
            log.warn('Nothing to update!')
            return

        # Validates updates against the remote ticket details.
        full_ticket = self._get_ticket(full=True)
        if args.start and full_ticket['current_workflow_step'] != (
                str(TicketWorkflowStep.NEW)):
            user_status = '{} - {}'.format(
                full_ticket['current_workflow_step'], full_ticket['status'])
            raise QueueingTicketError(
                'Ticked #{} has already started ({}) and cannot be '
                're-started manually.'.format(self.ticket_id, user_status))

        # Process given sources, if any is given.
        if args.sources:
            self._parse_sources(args)
            # Check suite/series coherence.
            if full_ticket['series'] != self.suite.split("-")[0]:
                raise MismatchSuitesError(
                    'Submitted sources are not targeted to ticket suite '
                    '({})'.format(full_ticket['series']))
            # Check sources/subtickets uniqueness.
            remote_sources = [
                (s['source_package_upload']['sourcepackage']['name'],
                 s['source_package_upload']['version'])
                for s in full_ticket.get('subticket', [])
                if s.get('source_package_upload') is not None
            ]
            for sub in self.subtickets.itervalues():
                if (sub.sourcepackage, sub.version) in remote_sources:
                    raise SourceAlreadyUploadedError(
                        'Submitted source {}_{} was already uploaded to '
                        'ticket #{}'.format(
                            sub.sourcepackage, sub.version, self.ticket_id))
            # Request user confirmation for adding sources.
            print('You are about to include the following new source(s) '
                  'to the ticket #{} ({}):'.format(
                      self.ticket_id, args.title or full_ticket['title']))
            for sub in self.subtickets.itervalues():
                print('  * {}_{}'.format(sub.sourcepackage, sub.version))
            confirmed = utils.get_user_confirmation(default='y')
            if not confirmed:
                log.warn('Update dismissed.')
                return
            # Creates subtickets for the processed sources.
            self._process_subtickets()

        # Patch the ticket with the reviewed details.
        ticket_path = os.path.join(utils.TICKET_BASE, self.ticket_id, '')
        ticket_url = urlparse.urljoin(utils.CI_URL, ticket_path)
        utils.patch(ticket_url, data=updates)

        print('Ticket #{} updated.'.format(self.ticket_id))

    def show(self, args):
        """Prints the context ticket details.

        The results is encoded as 'UTF-8' for wider terminal compatibility.
        """
        self.ticket_id = args.id
        ticket = self._get_ticket()
        if ticket['current_workflow_step'] == str(TicketWorkflowStep.FAILED):
            status_line = '{}'.format(ticket['status'])
        else:
            status_line = 'Current landing step: {} - {}'.format(
                ticket['current_workflow_step'], ticket['status'])
        details = [
            'Ticket {uuid} - {title}',
            '{status_line}',
            'Description: {description}',
            'Owner: {owner}',
            'Created: {created}',
            'Updated: {updated}',
            'Bug #: {bug_id}',
        ]
        ticket['status_line'] = status_line
        details_text = '\n'.join(details).format(**ticket)
        print(details_text.encode('utf-8'))

    def list(self, args):
        """Prints a list of tickets.

        Considers the 'limit' & 'offset' arguments if given, otherwise
        prints the most recent batch (up to 20 records).
        """
        if args.offset is None:
            num_tickets = self._get_tickets(limit=1)['meta']['total_count']
            offset = int(num_tickets) - args.limit
            tickets = self._get_tickets(limit=args.limit, offset=offset)
        else:
            tickets = self._get_tickets(limit=args.limit, offset=args.offset)

        for ticket in tickets['objects']:
            line = ('Ticket {uuid}: {title} is {current_workflow_step} '
                    '({status})').format(**ticket)
            print(line.encode('utf-8'))


def new_ticket(args):
    """Script action for creating a ticket."""
    ticket = Ticket()
    ticket.create(args)
    return ticket


def update_ticket(args):
    """Script action for updating a ticket."""
    ticket = Ticket()
    ticket.update(args)
    return ticket


def ticket_status(args):
    """Script action for listing tickets or showing one."""
    ticket = Ticket()
    if args.id:
        ticket.show(args)
    else:
        ticket.list(args)
    return ticket
