#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json

from ci_utils import amqp_utils
from ci_utils.data_store import DataStoreException
from coverageextractor import (
    CoverageExtractorException,
    extract_coverage_data,
    retry,
)

from nfss_client import NfssClient, NfssClientError

SUCCESS_TEMPLATE = u'Deposited coverage.xml into swift for ticket %s.'
MISSING_PARAMS_TEMPLATE = (u'Failed to handle request: missing coverage.xml '
                           'filename for ticket %s')
PARSE_FAILURE_TEMPLATE = u'Failed to parse coverage.xml for ticket %s: %s'
SWIFT_GET_FAILURE_TEMPLATE = (u'Failed to retrieve coverage.xml from swift '
                              'for ticket %s: %s')
SWIFT_PUT_FAILURE_TEMPLATE = (u'Failed to put coverage.xml into swift for '
                              'ticket %s: %s')
JSON_ENCODING_FAILURE_TEMPLATE = (u'Failed to encode JSON for '
                                  'ticket %s: %s')
NFSS_CLIENT_ERROR_TEMPLATE = u'Failed to POST to NFSS for ticket %s: %s'


def generate_and_put_artifact(
        data_store,
        name='coverage-extractor.output.log',
        type='LOGS',
        message=None,
        reference=None):
    """Return a uci-artifact generated and deposited in swift.

    NOTE that if no reference is supplied, the artifact is deposited
    into swift and the resulting link is reported in the returned
    artifact; provided for convenience in logging.

    :param data_store: DataStore into which to put
    :param name: filename of artifact in swift
    :param type: 'LOGS' for logs or 'RESULTS' for coverage.xml
    :param message: a log message
    :param reference: swift link; supplied upon deposit if not specified
    """
    artifact = {
        'name': name,
        'type': type,
    }
    if message is not None:
        artifact['message'] = message
    if reference is None:
        reference = data_store.put_file(name, str(artifact), 'text/plain')
    artifact['reference'] = reference
    return artifact


class ExtractionFailed(Exception):
    """Return a dict upon operation failure.."""

    def __init__(self, return_tuple):
        self.return_tuple = return_tuple


class Extraction(object):

    def __init__(self, ticket_id, logger, data_store, nfss_config_path=None):
        self.ticket_id = ticket_id
        self.logger = logger
        self.data_store = data_store
        # TODO: possibly we want to own the NFSS client here instead?
        self.nfss_config_path = nfss_config_path
        self.result = dict(artifacts=[])

    def try_get_project_and_version(self, params):
        """Extract project name and version from params.

        :param params: ticket system handle_request calling params
        """
        # FIXME: fginther, where to find these?
        return 'libfibonacci', '1.0-0ubuntu1'

    def try_get_coverage_xml_path(self, params):
        """Extract coverage.xml path from dict of params.

        :param params: dict, hopefully with 'coverage_xml_path' as key
        """
        try:
            return params['coverage_xml_path']
        except KeyError:
            self.report_failed(
                MISSING_PARAMS_TEMPLATE,
                self.ticket_id
            )

    def try_get_coverage_xml(self, coverage_xml_path):
        """Get coverage.xml at given path from swift.

        :param coverage_xml_path: swift path from which to get
        """
        try:
            @retry(DataStoreException, logger=self.logger)
            def get_build_log_actual():
                return self.data_store.get_file(coverage_xml_path)
            return get_build_log_actual()
        except DataStoreException as e:
            self.report_failed(
                SWIFT_GET_FAILURE_TEMPLATE,
                self.ticket_id,
                e.message,
            )

    def try_extract_coverage_data(self, coverage_xml):
        """Get line and branch percentages from coverage.xml.

        :param coverage_xml: coverage_xml string
        """
        try:
            return extract_coverage_data(coverage_xml)
        except CoverageExtractorException as e:
            self.report_failed(
                PARSE_FAILURE_TEMPLATE,
                self.ticket_id,
                e.message,
            )

    def try_compose_nfss_json(
            self,
            name,
            version,
            line_coverage,
            branch_coverage):
        """Compose a JSON string of data for submission to NFSS.

        :param name: package name
        :param version: package version string
        :param line_coverage: line coverage (float)
        :param branch_coverage: branch coverage (float)
        """
        try:
            return json.dumps({
                'name': name,
                'version': version,
                'line_coverage': line_coverage,
                'branch_coverage': branch_coverage,
            })
        except ValueError as e:
            # This is pretty unlikely.
            self.report_failed(
                JSON_ENCODING_FAILURE_TEMPLATE,
                self.ticket_id,
                e.message
            )

    def try_post_data_to_nfss(self, json, test_name, project_name='coverage'):
        """Post a JSON string of coverage data to NFSS.

        :param json: coverage data JSON to post to NFSS
        """
        try:
            nfss_client = NfssClient(self.nfss_config_path)
            nfss_client.post_data(json, test_name, project_name)
        except NfssClientError as e:
            self.report_failed(
                NFSS_CLIENT_ERROR_TEMPLATE,
                self.ticket_id,
                e.message
            )

    def report_failed(self, format_str, *str_args):
        """Convenience function to raise and log a failiure.

        :param format_str: template to log
        :str_args str_args: items with which to format format_str
        """
        message = format_str % str_args
        self.logger.error(message)
        self.result['message'] = message
        self.result['artifacts'].append(
            generate_and_put_artifact(self.data_store, message=message)
        )
        raise ExtractionFailed(
            (amqp_utils.progress_failed, self.result)
        )

    def report_succeeded(self, ticket_id):
        """Deposit a result artifact and return a log artifact.

        :param ticket_id: ticket for which to report retrieval success
        """
        message = SUCCESS_TEMPLATE % ticket_id
        # NOTE that logging compels us to pass params to info (can't
        # reuse message declared above)
        self.logger.info(SUCCESS_TEMPLATE, ticket_id)
        self.result['artifacts'].append(
            generate_and_put_artifact(
                self.data_store,
                message=message
            )
        )
        self.result['message'] = message
        return amqp_utils.progress_completed, self.result
