# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from coverageextractor import (
    extract_coverage_data,
    CoverageExtractorException,
)
from coverageextractor.tests import vanilla_coverage_xml


# here we refuse to close our <sources> tag :|
invalid_xml_coverage_xml = u"""<?xml version="1.0" ?>
<!DOCTYPE coverage
  SYSTEM 'http://cobertura.sourceforge.net/xml/coverage-03.dtd'>
<coverage branch-rate="1.0" line-rate="1.0" timestamp="1410387089" version="gcovr 2.4 (r2774)">
	<sources>
		<source>/tmp/tmpz8yrh4tz/dummy</source>
	<packages>
		<package branch-rate="1.0" complexity="0.0" line-rate="1.0" name="">
			<classes>
				<class branch-rate="1.0" complexity="0.0" filename="fibonacci.cpp" line-rate="1.0" name="fibonacci_cpp">
					<lines>
						<line branch="false" hits="10" number="4"/>
						<line branch="true" condition-coverage="100% (2/2)" hits="10" number="6">
							<conditions>
								<condition coverage="100%" number="0" type="jump"/>
							</conditions>
						</line>
						<line branch="false" hits="3" number="7"/>
						<line branch="true" condition-coverage="100% (2/2)" hits="7" number="8">
							<conditions>
								<condition coverage="100%" number="0" type="jump"/>
							</conditions>
						</line>
						<line branch="false" hits="4" number="9"/>
						<line branch="false" hits="3" number="10"/>
					</lines>
				</class>
			</classes>
		</package>
	</packages>
</coverage>
"""  # NOQA

# NOTE a *baseball* dingbat in sources filepath
unicode_coverage_xml = unicode("""<?xml version="1.0" ?>
<!DOCTYPE coverage
  SYSTEM 'http://cobertura.sourceforge.net/xml/coverage-03.dtd'>
<coverage branch-rate="1.0" line-rate="1.0" timestamp="1410387089" version="gcovr 2.4 (r2774)">
	<sources>
		<source>/tmp/\u26BE/dummy</source>
	</sources>
	<packages>
		<package branch-rate="1.0" complexity="0.0" line-rate="1.0" name="">
			<classes>
				<class branch-rate="1.0" complexity="0.0" filename="fibonacci.cpp" line-rate="1.0" name="fibonacci_cpp">
					<lines>
						<line branch="false" hits="10" number="4"/>
						<line branch="true" condition-coverage="100% (2/2)" hits="10" number="6">
							<conditions>
								<condition coverage="100%" number="0" type="jump"/>
							</conditions>
						</line>
						<line branch="false" hits="3" number="7"/>
						<line branch="true" condition-coverage="100% (2/2)" hits="7" number="8">
							<conditions>
								<condition coverage="100%" number="0" type="jump"/>
							</conditions>
						</line>
						<line branch="false" hits="4" number="9"/>
						<line branch="false" hits="3" number="10"/>
					</lines>
				</class>
			</classes>
		</package>
	</packages>
</coverage>
""")  # NOQA

# we'll almost certainly never see garbled coverage data but hey
garbled_numbers_coverage_xml = u"""<?xml version="1.0" ?>
<!DOCTYPE coverage
  SYSTEM 'http://cobertura.sourceforge.net/xml/coverage-03.dtd'>
<coverage branch-rate="garblegarble" line-rate="1.0" timestamp="1410387089" version="gcovr 2.4 (r2774)">
	<sources>
		<source>/tmp/tmpz8yrh4tz/dummy</source>
        </sources>
	<packages>
		<package branch-rate="1.0" complexity="0.0" line-rate="1.0" name="">
			<classes>
				<class branch-rate="1.0" complexity="0.0" filename="fibonacci.cpp" line-rate="1.0" name="fibonacci_cpp">
					<lines>
						<line branch="false" hits="10" number="4"/>
						<line branch="true" condition-coverage="100% (2/2)" hits="10" number="6">
							<conditions>
								<condition coverage="100%" number="0" type="jump"/>
							</conditions>
						</line>
						<line branch="false" hits="3" number="7"/>
						<line branch="true" condition-coverage="100% (2/2)" hits="7" number="8">
							<conditions>
								<condition coverage="100%" number="0" type="jump"/>
							</conditions>
						</line>
						<line branch="false" hits="4" number="9"/>
						<line branch="false" hits="3" number="10"/>
					</lines>
				</class>
			</classes>
		</package>
	</packages>
</coverage>
"""  # NOQA


class ExtractCoverageDataTestCase(unittest.TestCase):

    def test_extract_coverage_data(self):
        """Extract coverage data from a vanilla build log bytestring."""
        line_coverage, branch_coverage = extract_coverage_data(
            vanilla_coverage_xml
        )
        self.assertAlmostEqual(1.0, line_coverage)
        self.assertAlmostEqual(1.0, branch_coverage)

    def test_invalid_xml_raises(self):
        """We raise an exception if we find invalid XML."""
        self.assertRaises(
            CoverageExtractorException,
            extract_coverage_data,
            invalid_xml_coverage_xml
        )

    def test_garbled_numbers_raise(self):
        """We raise an exception if coverage numbers don't float."""
        self.assertRaises(
            CoverageExtractorException,
            extract_coverage_data,
            garbled_numbers_coverage_xml
        )

    def test_unicode_parses(self):
        """Extract coverage data from a build log with unicode characters."""
        line_coverage, branch_coverage = extract_coverage_data(
            unicode_coverage_xml
        )
        self.assertAlmostEqual(1.0, line_coverage)
        self.assertAlmostEqual(1.0, branch_coverage)
