#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from mock import patch, Mock

from ci_utils.data_store import DataStoreException
from coverageextractor import CoverageExtractorException
from coverageextractor.nfss_client import NfssClient, NfssClientError
from coverageextractor.tests import (
    CoverageExtractorTestCase,
    vanilla_coverage_xml,
)
from coverageextractor.extraction import (
    Extraction,
    ExtractionFailed,
    PARSE_FAILURE_TEMPLATE,
    SWIFT_GET_FAILURE_TEMPLATE,
    JSON_ENCODING_FAILURE_TEMPLATE,
    NFSS_CLIENT_ERROR_TEMPLATE,
    SUCCESS_TEMPLATE,
)


class ExtractionTestCase(CoverageExtractorTestCase):

    def test_try_get_coverage_xml_path(self):
        """We get the coverage.xml path from params."""
        with self.patch_create_data_store() as fake_data_store:
            params = {'coverage_xml_path': 'coverage.xml'}
            ext = Extraction('fake-ticket-id', 'fake-logger', fake_data_store)
            self.assertEqual(
                'coverage.xml',
                ext.try_get_coverage_xml_path(params),
            )

    @patch.object(Extraction, 'report_failed')
    def test_extract_coverage_xml_path_failure(self, report_failed):
        """Failure to get the coverage.xml path from params fails."""
        with self.patch_create_data_store() as fake_data_store:
            malevolent_params = {'gibberish': 'coverage.xml'}
            ext = Extraction('fake-ticket-id', 'fake-logger', fake_data_store)
            ext.try_get_coverage_xml_path(malevolent_params)
        report_failed.assert_called()

    def test_try_get_coverage_xml(self):
        """Retrieve our coverage.xml from swift."""
        with self.patch_create_data_store() as fake_data_store:
            fake_data_store.get_file.return_value = vanilla_coverage_xml
            ext = Extraction('fake-ticket-id', 'fake-logger', fake_data_store)
            self.assertEqual(
                vanilla_coverage_xml,
                ext.try_get_coverage_xml('coverage.xml'),
            )

    def test_try_get_coverage_xml_retries(self):
        """Retry if swift fails to yield our coverage.xml."""
        fake_logger = Mock(spec_set=logging.Logger)
        with self.patch_create_data_store() as fake_data_store:
            ext = Extraction('fake-ticket-id', fake_logger, fake_data_store)
            fake_data_store.get_file.side_effect = [
                DataStoreException("No log!"),
                DataStoreException("Still no log!"),
                vanilla_coverage_xml
            ]
            with patch('coverageextractor.time.sleep'):  # for the impatient
                self.assertEqual(
                    vanilla_coverage_xml,
                    ext.try_get_coverage_xml('coverage.xml')
                )
        self.assertEqual(3, fake_data_store.get_file.call_count)
        fake_logger.warning.assert_called()

    @patch.object(Extraction, 'report_failed')
    def test_try_get_coverage_xml_raises(self, report_failed):
        """Return failure if swift fails to yield our coverage.xml."""
        fake_logger = Mock(spec_set=logging.Logger)
        with self.patch_create_data_store() as fake_data_store:
            ext = Extraction('fake-ticket-id', fake_logger, fake_data_store)
            fake_data_store.get_file.side_effect = DataStoreException(
                'Drool...'
            )
            with patch('coverageextractor.time.sleep'):  # for the impatient
                ext.try_get_coverage_xml('coverage.xml')
        report_failed.assert_called_with(
            SWIFT_GET_FAILURE_TEMPLATE,
            'fake-ticket-id',
            'Drool...',
        )

    def test_try_extract_coverage_data(self):
        """Extract coverage data from coverage.xml."""
        ext = Extraction('fake-ticket-id', 'fake-logger', 'fake-data-store')
        line_coverage, branch_coverage = ext.try_extract_coverage_data(
            vanilla_coverage_xml
        )
        self.assertAlmostEqual(1.0, line_coverage)
        self.assertAlmostEqual(1.0, branch_coverage)

    @patch.object(Extraction, 'report_failed')
    @patch('coverageextractor.extraction.extract_coverage_data')
    def test_try_extract_coverage_data_raises(
            self,
            extract_coverage_data,
            report_failed):
        """Report failure when we fail to extract coverage data."""
        extract_coverage_data.side_effect = CoverageExtractorException(
            "No data for you!"
        )
        ext = Extraction('fake-ticket-id', 'fake-logger', 'fake-data-store')
        ext.try_extract_coverage_data(vanilla_coverage_xml)
        report_failed.assert_called_with(
            PARSE_FAILURE_TEMPLATE,
            'fake-ticket-id',
            "No data for you!",
        )

    def test_compose_nfss_json(self):
        """Dump a string of data to put to NFSS."""
        ext = Extraction('fake-ticket-id', 'fake-logger', 'fake-data-store')
        json = ext.try_compose_nfss_json(
            'fake-package-name',
            '0.1.PBJ.BBQ-0ubuntu1',
            1.0,
            0.572,
        )
        self.assertEqual(
            '{"version": "0.1.PBJ.BBQ-0ubuntu1", "name": "fake-package-name", "line_coverage": 1.0, "branch_coverage": 0.572}',  # NOQA
            json
        )

    @patch.object(Extraction, 'report_failed')
    @patch('coverageextractor.extraction.json.dumps')
    def test_compose_nfss_json_raises(self, dumps, report_failed):
        """Handle the error from un-jsonnable params.

        This is pretty unlikely.
        """
        dumps.side_effect = ValueError('This is pretty unlikely.')
        ext = Extraction('fake-ticket-id', 'fake-logger', 'fake-data-store')
        ext.try_compose_nfss_json(
            'fake-package-name',
            '0.1.PBJ.BBQ-0ubuntu1',
            1.0,
            0.572,
        )
        report_failed.assert_called_with(
            JSON_ENCODING_FAILURE_TEMPLATE,
            'fake-ticket-id',
            "This is pretty unlikely.",
        )

    @patch.object(NfssClient, 'post_data')
    def test_try_post_data_to_nfss(self, post_data):
        """Post JSON result to NFSS."""
        ext = Extraction('fake-ticket-id', 'fake-logger', 'fake-data-store')
        with patch.object(NfssClient, '__init__', new=Mock(return_value=None)):
            ext.try_post_data_to_nfss({'foo': 42}, 'libfibonacci', 'coverage')
        post_data.assert_called_with({'foo': 42}, 'libfibonacci', 'coverage')

    @patch.object(Extraction, 'report_failed')
    @patch.object(NfssClient, 'post_data')
    def test_nfss_doesnt_love_us(self, post_data, report_failed):
        """Bad post to NFSS reports failure."""
        ext = Extraction('fake-ticket-id', 'fake-logger', 'fake-data-store')
        post_data.side_effect = NfssClientError('NFSS luvs u <3 .')
        with patch.object(NfssClient, '__init__', new=Mock(return_value=None)):
            ext.try_post_data_to_nfss({'foo': 42}, 'libfibonacci', 'coverage')
        report_failed.assert_called_with(
            NFSS_CLIENT_ERROR_TEMPLATE,
            'fake-ticket-id',
            'NFSS luvs u <3 .',
        )

    def test_report_failed_raises(self):
        """Internal catch-all handler raises ExtractionFailed."""
        with self.patch_create_data_store() as fake_data_store:
            fake_logger = Mock(spec_set=logging.Logger)
            ext = Extraction('fake-ticket-id', fake_logger, fake_data_store)
            self.assertRaises(
                ExtractionFailed,
                lambda: ext.report_failed("%s %s", 'foo', 'bar')
            )
        fake_logger.error.assert_called_with("foo bar")

    @patch('coverageextractor.extraction.generate_and_put_artifact')
    def test_report_succeeded(self, generate_and_put_artifact):
        """Report success to swift and logging."""
        generate_and_put_artifact.side_effect = ['coverage.xml',
                                                 'success-message']
        with self.patch_create_data_store() as fake_data_store:
            fake_logger = Mock(spec_set=logging.Logger)
            ext = Extraction('fake-ticket-id', fake_logger, fake_data_store)
            ext.report_succeeded('fake-ticket-id')
        fake_logger.info.assert_called_with(SUCCESS_TEMPLATE, 'fake-ticket-id')
        generate_and_put_artifact.assert_any_call(
            fake_data_store,
            message=(SUCCESS_TEMPLATE % 'fake-ticket-id'),
        )
        self.assertEqual(1, generate_and_put_artifact.call_count)
