# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os

from mock import Mock, patch

from ci_utils import amqp_utils
from coverageextractor.extraction import Extraction
from coverageextractor.tests import CoverageExtractorTestCase
from coverageextractor.extraction import (
    MISSING_PARAMS_TEMPLATE,
    SUCCESS_TEMPLATE,
)
from coverageextractor.run_worker import CoverageExtractorWorker


class HandleRequestTestCase(CoverageExtractorTestCase):

    @patch.object(Extraction, 'try_post_data_to_nfss')
    def test_vanilla_handle_request(self, try_post_data_to_nfss):
        """Large integration test for coverage extraction.

        NOTE that we do patch the NFSS call.
        """
        fake_logger = Mock(spec_set=logging.Logger)
        with self.patch_create_data_store():
            self.put_fake_coverage_xml(self.fake_data_store)
            worker = CoverageExtractorWorker()
            fake_params = {'coverage_xml_path': 'coverage.xml',
                           'ticket_id': 'fake_ticket_id'}
            amqp_result, result_data = worker.handle_request(
                fake_params,
                fake_logger,
            )
        self.assertEqual(amqp_utils.progress_completed, amqp_result)
        self.assertEqual(
            SUCCESS_TEMPLATE % 'fake_ticket_id',
            result_data['message'],
        )
        fake_logger.info.assert_called_with(
            SUCCESS_TEMPLATE, 'fake_ticket_id'
        )
        self.assertEqual(1, len(result_data['artifacts']))
        # find a log message as our only artifact
        log_artifact = result_data['artifacts'][0]
        self.assertEqual(
            SUCCESS_TEMPLATE % 'fake_ticket_id',
            log_artifact['message'],
        )
        self.assertEqual('LOGS', log_artifact['type'])
        self.assertEqual(
            'file://' + os.path.join(
                os.path.abspath(os.curdir),
                'fake_ticket_id/coverage-extractor.output.log'),
            log_artifact['reference']
        )

    def test_exception_returns_failure(self):
        """An exception in our imperatives returns a failure.

        Actuate a KeyError in try_extract_coverage_xml_path, inspect our
        results to make sure we're logging, depositing a log in swift.
        """
        fake_logger = Mock(spec_set=logging.Logger)
        with self.patch_create_data_store() as fake_data_store:
            self.put_fake_coverage_xml(fake_data_store)
            worker = CoverageExtractorWorker()
            # actuate a KeyError in try_extract_coverage_xml_path
            malevolent_params = {'gibberish': 'coverage.xml',
                                 'ticket_id': 'fake_ticket_id'}
            amqp_result, result_data = worker.handle_request(
                malevolent_params,
                fake_logger
            )
        self.assertEqual(amqp_utils.progress_failed, amqp_result)
        self.assertEqual(
            MISSING_PARAMS_TEMPLATE % 'fake_ticket_id',
            result_data['message']
        )
        self.assertEqual(1, len(result_data['artifacts']))
        artifact = result_data['artifacts'][0]
        self.assertEqual(
            'file://' + os.path.join(
                os.path.abspath(os.curdir),
                'fake_ticket_id/coverage-extractor.output.log'),
            artifact['reference'])
        self.assertEqual(
            MISSING_PARAMS_TEMPLATE % 'fake_ticket_id',
            result_data['message']
        )
        self.assertEqual('LOGS', artifact['type'])
