#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from mock import Mock, patch

from coverageextractor.nfss_client import (
    NFSS_CONFIG_PATH,
    NfssClient,
    NfssClientError,
    parse_nfss_config,
)


class ParseNfssConfigTestCase(unittest.TestCase):

    @patch('coverageextractor.nfss_client.ConfigParser')
    def test_parse(self, config_parser):
        """We can successfully parse an NFSS config."""
        nfss_config_dict = {
            'client_access_key': 'fake-client-acccess-key',
            'resource_owner_key': 'fake-resource-owner-key',
            'resource_owner_secret': 'fake-resource-owner-secret',
            'backend': 'fake-backend',
        }
        # yes these are weird due to ConfigParser backwardness
        config_parser.return_value.read = Mock()
        # "dance like no-one is watching"
        config_parser.return_value.__getitem__ = Mock(
            return_value=nfss_config_dict
        )
        self.assertEqual(
            nfss_config_dict,
            parse_nfss_config('fake-config-path')
        )

    @patch('coverageextractor.nfss_client.ConfigParser')
    def test_no_nfss_section_raises(self, config_parser):
        """No 'nfss' section in config raises."""
        config_parser.return_value.read = Mock()
        config_parser.return_value.__getitem__ = Mock(
            side_effect=KeyError('client_access_key')
        )
        with self.assertRaises(KeyError):
            parse_nfss_config('fake-config-path')

    @patch('coverageextractor.nfss_client.ConfigParser')
    def test_missing_key_raises(self, config_parser):
        """A missing key in the config raises."""
        nfss_config_dict = {}
        config_parser.return_value.read = Mock()
        config_parser.return_value.__getitem__ = Mock(
            return_value=nfss_config_dict
        )
        with self.assertRaises(KeyError):
            parse_nfss_config('fake-config-path')


class NfssClientTest(unittest.TestCase):

    @patch('coverageextractor.nfss_client.OAuth1Session')
    @patch('coverageextractor.nfss_client.parse_nfss_config')
    def test_init(self, parse_nfss_config, oauth_session):
        """We can create a NfssClient."""
        parse_nfss_config.return_value = {
            'client_access_key': 'fake-client-access-key',
            'resource_owner_key': 'fake-resource-owner-key',
            'resource_owner_secret': 'fake-resource-owner-secret'
        }
        NfssClient('fake-config-path')
        oauth_session.assert_called_with(
            'fake-client-access-key',
            resource_owner_key='fake-resource-owner-key',
            resource_owner_secret='fake-resource-owner-secret',
        )

    @patch('coverageextractor.nfss_client.OAuth1Session')
    @patch('coverageextractor.nfss_client.parse_nfss_config')
    def test_no_config_path(self, parse_nfss_config, oauth_session):
        """Config path defaults if not passed in as param."""
        parse_nfss_config.return_value = {
            'client_access_key': 'fake-client-access-key',
            'resource_owner_key': 'fake-resource-owner-key',
            'resource_owner_secret': 'fake-resource-owner-secret'
        }
        nfss_client = NfssClient()
        oauth_session.assert_called_with(
            'fake-client-access-key',
            resource_owner_key='fake-resource-owner-key',
            resource_owner_secret='fake-resource-owner-secret',
        )
        self.assertEqual(NFSS_CONFIG_PATH, nfss_client.nfss_config_path)

    @patch('coverageextractor.nfss_client.parse_nfss_config')
    def test_init_raises(self, parse_nfss_config):
        """We can create a NfssClient."""
        parse_nfss_config.side_effect = KeyError('Mangled config!')
        with self.assertRaises(NfssClientError):
            NfssClient('fake-config-path')

    def test_post_data(self):
        """Vanilla POST to NFSS."""
        with patch.object(
                NfssClient,
                '__init__',
                new=Mock(return_value=None)):
            nfss_client = NfssClient()
            nfss_client.config = {'backend': 'http://NFSS.io'}
            nfss_client.session = Mock(
                # successful POST
                post=Mock(return_value=Mock(status_code=200)),
            )
            nfss_client.post_data(
                {'foo': 'bar'},
                'coverage',
                'libfibonacci',
            )

    def test_bad_post_data_raises(self):
        """Vanilla POST to NFSS."""
        with patch.object(
                NfssClient,
                '__init__',
                new=Mock(return_value=None)):
            nfss_client = NfssClient()
            nfss_client.config = {'backend': 'http://NFSS.io'}
            nfss_client.session = Mock(
                # successful POST
                post=Mock(return_value=Mock(status_code=500)),
            )
            with self.assertRaises(NfssClientError):
                nfss_client.post_data(
                    {'foo': 'bar'},
                    'coverage',
                    'libfibonacci',
                )
