# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from cStringIO import StringIO
import logging
import os
import unittest

from ucitests import (
    assertions,
    fixtures as uci_fixtures,
    scenarii,
)
from ucivms.tests import fixtures as vms_fixtures
from ci_utils import (
    amqp_utils,
    image_store,
    unit_config,
)
from ci_utils.testing import (
    features,
    fixtures,
)
from tstrun import (
    run_worker,
    testbed,
)


load_tests = scenarii.load_tests_with_scenarios


@features.requires(features.nova_creds)
@features.requires(features.nova_compute)
class TestWorkerHandleRequest(unittest.TestCase):

    def setUp(self):
        super(TestWorkerHandleRequest, self).setUp()
        # A logger for the test runner worker
        self.logger_stream = logging.StreamHandler(StringIO())
        self.logger_stream.setFormatter(logging.Formatter('%(message)s'))
        self.logger = logging.getLogger()
        self.logger.addHandler(self.logger_stream)
        # The fake rabbit input queue
        self.queue = fixtures.FakeTaskQueue('test')
        # The data store and uci-vms need disk access
        vms_fixtures.isolate_from_disk(self)
        self.ds_path = os.path.join(self.uniq_dir, 'data_store')
        # We don't need a config for the data store
        self.data_store = fixtures.FakeDataStore(self.ds_path)

        def create_data_store(component):
            return self.data_store

        self.ds_factory = create_data_store
        self.conf = unit_config.get_auth_config()

    def get_ppa_list(self):
        # We're testing only the test runner so there is no ppa assigned to the
        # ticket, only the master ppa is relevant.
        return [self.conf.get('master_ppa')]

    def get_image_id(self, series, arch):
        return image_store.uci_image_name('cloudimg', series, arch)

    def assertTestRun(self, expected_artifacts, params, worker=None):
        if worker is None:
            worker = run_worker.TestRunnerWorker(self.ds_factory)
        worker = run_worker.TestRunnerWorker(self.ds_factory)
        self.queue.publish(params)
        returns = self.queue.subscribe(worker.handle_request, self.logger)
        assertions.assertLength(self, 1, returns)
        self.assertTrue(self.queue.is_empty())
        self.assertEquals(expected_artifacts, self.data_store.list_files())
        return returns

    def test_two_packages_two_testbeds(self):
        # We use precise where no tests are defined for libpng nor adduser to
        # speed things up a bit without reducing coverage. (~244s instead of
        # 560~ on a single sample (YMMV, high variations there)).
        series = 'precise'
        architecture = 'amd64'
        params = dict(ticket_id=series, progress_trigger='progress',
                      series=series,
                      architecture=architecture,
                      image_id=self.get_image_id(series, architecture),
                      ppa_list=self.get_ppa_list(),
                      package_list=['libpng', 'adduser'])
        returns = self.assertTestRun(
            ['adduser-summary.log', 'adduser-syslog.log',
             'adduser-testbed-cloud-init.log',
             'adduser.log',
             'adduser.subunit',
             'libpng-summary.log', 'libpng-syslog.log',
             'libpng-testbed-cloud-init.log',
             'libpng.log',
             'libpng.subunit'],
            params)
        (retcode, results) = returns[0]
        self.assertEqual(amqp_utils.progress_failed, retcode)
        self.assertEquals(
            '*                    SKIP no tests in this package\n',
            self.data_store.get_file('libpng-summary.log'))
        self.assertEquals(
            '*                    SKIP no tests in this package\n',
            self.data_store.get_file('adduser-summary.log'))

    def test_teardown_dont_break_test_run(self):
        series = 'trusty'
        architecture = 'amd64'
        params = dict(ticket_id=series, progress_trigger='progress',
                      series=series,
                      architecture=architecture,
                      image_id=self.get_image_id(series, architecture),
                      ppa_list=self.get_ppa_list(),
                      package_list=['libpng'])
        worker = run_worker.TestRunnerWorker(self.ds_factory)

        def broken_teardown(test_bed):
            self.addCleanup(test_bed.nova.delete_server, test_bed.instance.id)
            if test_bed.floating_ip is not None:
                self.addCleanup(test_bed.nova.delete_floating_ip,
                                test_bed.floating_ip)
            raise AssertionError('Boom !')

        uci_fixtures.patch(self, testbed.TestBed, 'teardown', broken_teardown)
        returns = self.assertTestRun(
            ['libpng-summary.log', 'libpng-syslog.log',
             'libpng-testbed-cloud-init.log', 'libpng.log',
             'libpng.subunit'],
            params, worker)
        (retcode, results) = returns[0]
        self.assertEqual(amqp_utils.progress_completed, retcode)


@features.requires(features.nova_creds)
@features.requires(features.nova_compute)
class TestWorkerHandleRequestFor(unittest.TestCase):

    scenarios = scenarii.multiply_scenarios(
        # series
        ([('precise', dict(series='precise', result='skip')),
          ('trusty', dict(series='trusty', result='pass')),
          ('utopic', dict(series='utopic', result='pass'))]),
        # architectures
        ([('amd64', dict(arch='amd64')), ('i386', dict(arch='i386'))]))

    def setUp(self):
        super(TestWorkerHandleRequestFor, self).setUp()
        # A logger for the test runner worker
        self.logger_stream = logging.StreamHandler(StringIO())
        self.logger_stream.setFormatter(logging.Formatter('%(message)s'))
        self.logger = logging.getLogger()
        self.logger.addHandler(self.logger_stream)
        # We want the log to be displayed if a test fails
        self.display_log = True
        # If the test doesn't set display_log to False (upon completion), the
        # log is displayed

        # We need to delay the log acquisition until we attempt to display it
        def delayed_display_log():
            self.assertFalse(self.display_log,
                             msg='test logger: {}'.format(
                                 self.logger_stream.stream.getvalue()))
        self.addCleanup(delayed_display_log)
        # The fake rabbit input queue
        self.queue = fixtures.FakeTaskQueue('test')
        # The data store and uci-vms need disk access
        vms_fixtures.isolate_from_disk(self)
        self.ds_path = os.path.join(self.uniq_dir, 'data_store')
        # We don't need a config for the data store
        self.data_store = fixtures.FakeDataStore(self.ds_path)

        def create_data_store(component):
            return self.data_store

        self.ds_factory = create_data_store
        self.conf = unit_config.get_auth_config()

    def get_ppa_list(self):
        # We're testing only the test runner so there is no ppa assigned to the
        # ticket, only the master ppa is relevant.
        return [self.conf.get('master_ppa')]

    def get_image_id(self):
        return image_store.uci_image_name('cloudimg', self.series, self.arch)

    def assertTestRun(self, params, worker=None):
        if worker is None:
            worker = run_worker.TestRunnerWorker(self.ds_factory)
        worker = run_worker.TestRunnerWorker(self.ds_factory)
        self.queue.publish(params)
        returns = self.queue.subscribe(worker.handle_request, self.logger)
        assertions.assertLength(self, 1, returns)
        self.assertTrue(self.queue.is_empty())
        return returns

    def test_handle_request(self):
        ticket_id = '{}-{}'.format(self.series, self.arch)
        params = dict(ticket_id=ticket_id, progress_trigger='progress',
                      series=self.series,
                      architecture=self.arch,
                      image_id=self.get_image_id(),
                      ppa_list=self.get_ppa_list(),
                      package_list=['libpng'])
        returns = self.assertTestRun(params)
        (retcode, results) = returns[0]
        if self.result == 'skip':
            e_retcode = amqp_utils.progress_failed
            e_summary = '*                    SKIP no tests in this package\n'
        elif self.result == 'pass':
            e_retcode = amqp_utils.progress_completed
            e_summary = 'build                PASS\n'
        self.assertEqual(e_retcode, retcode)
        self.assertEquals(['libpng-summary.log', 'libpng-syslog.log',
                           'libpng-testbed-cloud-init.log', 'libpng.log',
                           'libpng.subunit'],
                          self.data_store.list_files())
        # We need to check the content here as the worker will put an error
        # message there if it cannot get them from the testbed (so the user get
        # *some* feedback).
        self.assertEquals(e_summary,
                          self.data_store.get_file('libpng-summary.log'))
        # The test succeeded, we don't need the log
        self.display_log = False
