#!/usr/bin/env python
# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import imp
import mock
import os
import requests
import unittest
import urlparse

from ci_utils.testing import (
    get_test_file_path,
    TestCaseWithGnupg,
)
import deployers


HERE = os.path.abspath(os.path.dirname(__file__))
CLI_PATH = os.path.join(HERE, "../cli/ubuntu-ci")


class CliTest(deployers.DeployerTest, TestCaseWithGnupg):
    """Integration tests for CLI on a juju deployment."""

    def setUp(self):
        super(CliTest, self).setUp()
        self.url = 'http://{}:{}/'.format(
            *self.get_ip_and_port('ci-airline-webui-apache'))
        self.cli = imp.load_source("ubuntu_ci", CLI_PATH)

    def get_tickets_count(self):
        """Return the number of tickets currently available."""
        url = '/api/v1/ticket/'
        response = requests.get(urlparse.urljoin(self.url, url))
        response.raise_for_status()
        return response.json()['meta']['total_count']

    def get_tickets(self, offset=0):
        """Return a list of tickets (dict) from the API sorted by 'ID'."""
        url = '/api/v1/ticket/?offset={}'.format(offset)
        response = requests.get(urlparse.urljoin(self.url, url))
        response.raise_for_status()
        return response.json()['objects']

    def get_ticket(self, ticket_id, full=False):
        """Return the ticket (dict) from the API."""
        model = 'ticket'
        if full:
            model = 'fullticket'
        url = urlparse.urljoin(
            self.url, '/api/v1/{}/{}'.format(model, ticket_id))
        response = requests.get(url)
        response.raise_for_status()
        return response.json()

    def test_create_ticket(self):
        # A new ticket can be created using the CLI.
        num_tickets = self.get_tickets_count()
        changes_path = get_test_file_path('foobar_0.1-1_source.changes')
        args = [
            '--url', self.url,
            'create_ticket',
            '-t', 'Created via CLI',
            '-d', 'This is a cool new feature',
            '-b', '4321',
            '-o', 'someone@example.com',
            '-s', changes_path,
        ]
        self.cli.main(args)
        new_tickets = self.get_tickets(offset=num_tickets)
        [new_ticket] = new_tickets
        self.assertEqual(
            'Created via CLI', new_ticket['title'])
        self.assertEqual(
            'This is a cool new feature', new_ticket['description'])
        self.assertEqual(4321, new_ticket['bug_id'])
        self.assertEqual('someone@example.com', new_ticket['owner'])

    def test_update_ticket_attributes(self):
        # An existing ticket can be updated using the CLI.
        num_tickets = self.get_tickets_count()
        ticket_id = self.get_tickets(offset=num_tickets - 1)[0]['uuid']
        args = [
            '--url', self.url,
            'update_ticket',
            '-i', str(ticket_id),
            '-t', 'Updated via CLI',
            '-d', 'This is *still* a cool new feature',
            '-b', '1234',
            '-o', 'someone-else@example.com',
        ]
        self.cli.main(args)
        ticket = self.get_ticket(ticket_id)
        self.assertEqual(
            'Updated via CLI', ticket['title'])
        self.assertEqual(
            'This is *still* a cool new feature', ticket['description'])
        self.assertEqual(1234, ticket['bug_id'])
        self.assertEqual('someone-else@example.com', ticket['owner'])

    @mock.patch('ci_cli.utils.local_raw_input', return_value='')
    def test_update_ticket_new_sources(self, mocked_raw):
        # Adding new sources to an existing ticket.
        # The system will automatically identify and validate the changes
        # and if necessary restart the ticket processing.
        num_tickets = self.get_tickets_count()
        ticket_id = self.get_tickets(offset=num_tickets - 1)[0]['uuid']
        changes = get_test_file_path('barfoo_0.1-1_source.changes')
        args = [
            '--url', self.url,
            'update_ticket',
            '-i', str(ticket_id),
            '-s', changes,
        ]
        self.cli.main(args)
        ticket = self.get_ticket(ticket_id, full=True)
        self.assertEqual(
            ['barfoo', 'foobar'],
            sorted([s['source_package_upload']['sourcepackage']['name']
                    for s in ticket['subticket']]))


if __name__ == "__main__":
    unittest.main()
